from fastapi import Depends
from functools import lru_cache
from pydantic_settings import BaseSettings
from typing import Any, Callable, Set, List, Optional, Annotated
import logging
import octomy.access
import octomy.config
import octomy.db
import pprint
import asyncio

logger = logging.getLogger(__name__)


try:
	from octomy.batch import Processor
	#raise ImportError
except ImportError:
	import sys
	import types

	# Check if 'octomy' module exists, if not create it
	if 'octomy' in sys.modules:
		octomy = sys.modules['octomy']
	else:
		octomy = types.ModuleType('octomy')
		sys.modules['octomy'] = octomy

	# Check if 'batch' submodule exists within 'octomy', if not create it
	if hasattr(octomy, 'batch'):
		batch = octomy.batch
	else:
		batch = types.ModuleType('batch')
		sys.modules['octomy.batch'] = batch
		octomy.batch = batch

	# Define the Processor class within octomy.batch
	class Processor:
		def __init__(self):
			print("octomy.batch.Processor is not installed. Using dummy Processor.")

		def is_dummy(self):
			return True
	logger.info("Asimilated batch processor")
	octomy.batch.Processor = Processor


DEFAULT_ADMIN_PATH="overwatch"

def url_for(name:str) -> str:
	url = f"http://CONTEXT_URL_FOR/{name}"
	logger.info(f"Generated url '{url}'")
	return  url



#required_scopes=["overwatch.access.profile.view"]
#required_scopes: List[str]

class Context(BaseSettings):
	config: octomy.config.OctomyConfig | None
	access: octomy.access.AccessContext | None
	db: octomy.db.Database | None
	bp: octomy.batch.Processor | None
#	current_user: Optional[octomy.access.User] = octomy.access.get_current_user()
#	url_for:Callable[[str], str] = url_for # SUBSTITUTE WITH NATIVE FASTAPI request.url_for() INSTEAD
	admin_path:str = DEFAULT_ADMIN_PATH
	required_scopes:List[str] = list()
	optional_attributes:dict = dict()
	#@property
	#def url_for(self, key) -> str:
	#	return url_for(key)
	webroot: str = "/app/webroot"
	password_cost_factor:int = 12
	def set_optional_attribute(self, name, value):
		self.optional_attributes[name] = value
	def __getattr__(self, name):
		if name in self.optional_attributes:
			val = self.optional_attributes.get(name)
			logger.info(f"Referencing optional attribute '{name}': {val}")
			return val
		logger.error(f"Trying to referene missing Context attribute '{name}'")
		return None


def default_decorator_function(undecorated_context:Context):
	return undecorated_context

_dec_fun = default_decorator_function

def set_context_decorator(dec_fun):
	_dec_fun = dec_fun



def scoped_context(s:list):
	return Annotated[Context, Depends(context_dependency(required_scopes=s))]


_context_singleton:Context = None

#@lru_cache
def get_context(required_scopes: Optional[List[str]] = None) -> Context:
	global _context_singleton
	do_debug = False
	if not _context_singleton:
		# We expect a list to be valid
		required_scopes = required_scopes or list()
		config:octomy.config.OctomyConfig = octomy.config.get_config()
		if None == config:
			config_error = "Unknown error"
			logger.error(f"Could not get config: {config_error}")
			return None
		db:octomy.db.Database | None = None
		bp:octomy.batch.Processor | None = None
		access:octomy.access.AccessContext | None = None
		result = octomy.db.get_database(config)
		db, db_err = result
		if not db:
			logger.warning(f"Could not get db: {db_err}")
		else:
			bp = octomy.batch.Processor(config = config, dbc = db)
			if not bp:
				bp_error = "Unknown error"
				logger.warning(f"Could not get config: {bp_error}")
			else:
				access = octomy.access.get_access_context(required_scopes)
				if not access:
					access_error = "Unknown error"
					logger.warning(f"Could not get config: {access_error}")
				else:
					if do_debug:
						logger.info(f"AccessContext: {pprint.pformat(access)}")
		context = Context(
			  config = config
			, access = access
			, db = db
			, bp = bp
			, admin_path = config.get("app-admin-path", DEFAULT_ADMIN_PATH)
			, required_scopes = required_scopes
		)
		_dec_fun(context)
		_context_singleton = context
	return _context_singleton


# Helper dependency
def context_dependency(required_scopes: Optional[List[str]] = None):
	async def dependency():
		return get_context(required_scopes = required_scopes or list())
	return dependency




###########################################

'''
TODO: Fix access

from fastapi.openapi.models import APIKey, APIKeyIn
from fastapi.security.base import SecurityBase

class ContextScheme(SecurityBase):

	def __init__(
		  self
		, scopes: Annotated[
			Optional[List[str]],
			Doc(
				"""
				The seucirty scopes required by the endpoint that depends on this schema instance
				"""
			),
		] = None
		, needsAuth: Annotated[
		bool,
		Doc(
			"""
			Whether or not we need a user that is authenticated
			"""
		),
		] = True
		, tokenUrl: Annotated[
			str,
			Doc(
				"""
				The URL to obtain the OAuth2 token. This would be the *path operation*
				that has `OAuth2PasswordRequestForm` as a dependency.
				"""
			),
		] = "token"

	):
		# Magic to make fastapi openapi happy
		self.model: APIKey = APIKey(
			**{"in": APIKeyIn.query},  # type: ignore[arg-type]
			name = "octomy_auth_cookie",
			description = "octomy_auth_cookie",
		)
		self.scheme_name = self.__class__.__name__
		self.tokenUrl = tokenUrl
		self.scopes = scopes or list()
		logger.info(f"ContextScheme.init(scopes={pprint.pformat(self.scopes)}, tokenUrl={pprint.pformat(self.tokenUrl)})");

	async def __call__(self, request:Request, cookie: str = Depends (api_cookie)) -> Context:
		do_debug = True
		local_user, local_user_err = get_current_local_user(cookie)
		user = None
		if local_user:
			user = User()
			user.local_user = local_user
		else:
			logger.error(f"Could not get user: {local_user_err}")
			
		return await get_context(user)
		
# Convenience export
Depends = fastapi.Depends
'''





from bs4 import BeautifulSoup
import logging
import os
import jellyfish
import pprint


logger = logging.getLogger(__name__)


def file_to_list(fn):
	out=[]
	with open(fn, "r") as f:
		for line in f:
			out.append(line.strip())
	return out


def spread_keywords(keywords, w):
	adds = []
	for i in range(0, len(keywords)-w):
		add=[]
		for j in range(0, w):
			add.append(keywords[i+j])
		adds.append(" ".join(add))
	return adds


phonetic_identifier="🚡"
def phonify(keyword):
	return [
		  f"{phonetic_identifier}{keyword.lower()}"
		, f"{phonetic_identifier}{jellyfish.soundex(keyword)}"
		, f"{phonetic_identifier}{jellyfish.nysiis(keyword)}"
#		, f"{phonetic_identifier}{jellyfish.match_rating_codex(keyword)}"
		, f"{phonetic_identifier}{jellyfish.metaphone(keyword)}"
	#   , f"{phonetic_identifier}{jellyfish.porter_stem}"
	]
 
def decorate_keywords(keywords):
	phonetic = []
	for keyword in keywords:
		if keyword:
			phonetic += phonify(keyword)
	return phonetic

def stopword_filtered(keywords, stopwords):
	out=[]
	for keyword in keywords:
		if keyword not in stopwords:
			out.append(keyword)
	return out
 

def saturate_keywords(keywords, stopwords):
	filtered = stopword_filtered(keywords, stopwords)
	for i in range(2, 5):
		keywords += spread_keywords(keywords, i)
		keywords += spread_keywords(filtered, i)
	keywords += decorate_keywords(keywords)
	keywords += decorate_keywords(filtered)
	return keywords



class SearchEngine:
	def __init__(self):
		self.page_index={}
		self.page_text={}
		self.page_title={}
		self.page_text_index={}
		self.stopwords=file_to_list(f"{os.path.dirname(__file__)}/stopwords.txt")
		#logger.info("STOPWORDS:")
		#logger.info(", ".join(self.stopwords))
		#logger.info(f"INDEX: {pprint.pformat(self.page_index)}")

	def index_page(self, path):
		html, err = handler_raw(path)
		if err:
			logger.error(f"SKIPPING PAGE {path}")
			return False
		soup = BeautifulSoup(html)
		title = soup.title.get_text()
		page_title[path] = title
		text = soup.get_text()
		keywords = text.split()
		text = " ".join(keywords)
		page_text[path] = text
		keywords = saturate_keywords(keywords, self.stopwords)
		counts = {}
		for keyword in keywords:
			counts[keyword] = counts.get(keyword, 0) + 1
		#logger.info(f"Indexed {path} into:")
		#logger.info(pprint.pformat(text))
		#logger.info(pprint.pformat(counts))
		for keyword in keywords:
			count = counts.get(keyword, 1)
			pages = page_index.get(keyword, dict())
			pages[path] = count + pages.get(path, 0)
			page_index[keyword] = pages
		for keyword in keywords:
			pages = page_text_index.get(keyword, dict())
			hit=text.find(keyword)
			if -1 != hit:
				pages[path] = hit
			page_text_index[keyword] = pages
		return True
	
	def summarize(self, keyword, path, width=10):
		pages = page_text_index.get(keyword, dict())
		index = pages.get(path, None)
		if index:
			text = page_text.get(path, None)
			if text:
				s = index-width
				l = len(keyword)+width*2
				logger.info(f"s={s},l={l}")
				return text[s:l]
		return ""
	
	def find_in_index(self, q, max=10):
		if not q or ""==q:
			return ""
		ret = {}
		score = 0
		saturated_q = saturate_keywords([q], self.stopwords)
		logger.info(f"saturated q: {pprint.pformat(saturated_q)} ############################")
		for s in saturated_q:
			hits = page_index.get(s, dict())
			logger.info(f"hits for {s}: {pprint.pformat(hits)}")
			for path, count in hits.items():
				ret[path] = count + ret.get(path, 0)
		logger.info(f"Total: {pprint.pformat(ret)}")
		logger.info("")
		return ret
	
	
	
	def federated_search(self, q:str):
		ret = find_in_index(q)
		logger.info(f"HIT:{pprint.pformat(ret)}")
		out=[]
		for path, score in ret.items():
			hit = [score # Score
			, page_title.get(path, "Unknown") # Title
			, self.summarize(q, path) #Context
			, 6 # Beginning
			, 5 # length
			, path # Link URL
			]
			out.append(hit)



from google.auth.transport.requests import Request
from google.oauth2 import service_account
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.http import MediaFileUpload, MediaIoBaseUpload
import google.auth
import googleapiclient.discovery
import googleapiclient.errors
import io
import json
import logging
import os
import pprint
import requests
import socket


logger = logging.getLogger(__name__)

timeout_in_sec = 3
socket.setdefaulttimeout(timeout_in_sec)

folder_id = os.environ.get("GOOGLE_DRIVE_FOLDER_ID")
#google_username = "oktopus-development@oktopus-397222.iam.gserviceaccount.com"

FOLDER_MIMETYPE = "application/vnd.google-apps.folder"


class GoogleDrive:
	default_mimetype = 'application/octet-stream'

	def __init__(self, credentials_file_path = None):
		self.common_paramters = {
			"supportsAllDrives": True
		}
		self.search_paramters = {
			  "includeItemsFromAllDrives": True
			, "corpora": 'allDrives'
		}
		logger.info(f"GOOGLE_APPLICATION_CREDENTIALS: {os.environ.get('GOOGLE_APPLICATION_CREDENTIALS', 'not present')}")
		logger.info(f"GOOGLE_DRIVE_FOLDER_ID: {os.environ.get('GOOGLE_DRIVE_FOLDER_ID', 'not present')}")
		"""
		Load pre-authorized user credentials from the environment.
		TODO(developer) - See https://developers.google.com/identity
		for guides on implementing OAuth2 for the application.
		"""
		creds = None
		project_id = None
		if credentials_file_path:
			logger.info(f"Cred file specified, loading creds from there")
			drive_scopes = ['https://www.googleapis.com/auth/drive', 'https://www.googleapis.com/auth/drive.file']
			creds = service_account.Credentials.from_service_account_file(
										  filename = credentials_file_path, 
										  scopes = drive_scopes)
			"""
			with open(credentials_file_path, 'r') as credentials_file:
				credentials_data = json.load(credentials_file)
				logger.info(f"DATA:")
				logger.info(pprint.pformat(credentials_data))
				creds = google.auth.credentials.Credentials.from_service_account_info(credentials_data)
			"""
			project_id = creds.project_id
		else:
			logger.info(f"Cred file not specified, loading creds from environment")
			creds, project_id = google.auth.default()
		
		logger.info(f"creds='{pprint.pformat(creds)}', project_id='{pprint.pformat(project_id)}'")
		# create drive api client
		self.creds = creds
		self.service = googleapiclient.discovery.build('drive', 'v3', credentials = self.creds)
		logger.info('Initialized google drive api')
		#raise Exception("gd here")

	def search_files(self, q, fields=list(), do_debug = False):
		"""Search file in drive location
		"""
		fields.append('id')
		fields.append('name')
		fields = list(set(fields))
		files = []
		page_token = None
		if do_debug:
			logger.info(f" ## Searching for files with query=\"{q}\" and fields=[ {', '.join(fields)} ] on google drive")
		try:
			while True:
				# pylint: disable=maybe-no-member
				response = self.service.files().list(q = q,
												spaces = 'drive',
												fields = 'nextPageToken, '
													   f'files({",".join(fields)})',
												pageToken = page_token,
												**self.common_paramters,
												**self.search_paramters
												).execute()
				#for file in response.get('files', list()):
					# Process change
					#logger.info(F'Found file: {file.get("name")}, {file.get("id")}')
				files.extend(response.get('files', list()))
				page_token = response.get('nextPageToken', None)
				if page_token is None:
					break
	
		except googleapiclient.errors.HttpError as error:
			logger.error(f'An error occurred: {error}')
			files = None
	
		return files

	def recursive_search_worker(self, folder_id, q_files, q_folders, fields=list(), path=list(), exclude_dot_files = True, exclude_dot_folders = True, do_debug = False):
		"""Search files and folders in drive location (one recursive step)
		"""
		path = path or list()
		fields.append('id')
		fields.append('name')
		fields = list(set(fields))
		files = list()
		folders = list()
		
		sub_q_files =   f"'{folder_id}' in parents and mimeType != '{FOLDER_MIMETYPE}' and {q_files}"
		sub_q_folders = f"'{folder_id}' in parents and mimeType =  '{FOLDER_MIMETYPE}' and {q_folders}"
		# https://developers.google.com/drive/api/guides/ref-search-terms#operators
		
		if do_debug:
			logger.info(f" ## Searching recursively for files and folders under {folder_id}")
			logger.info(f"    q_files=\"{q_files}\"")
			logger.info(f"    q_folders=\"{q_folders}\"")
			logger.info(f"    sub_q_files=\"{sub_q_files}\"")
			logger.info(f"    sub_q_folders=\"{sub_q_folders}\"")
			logger.info(f"    fields=[ {', '.join(fields)} ] ")
			logger.info(f"    on google drive")
		try:
			page_token = None
			while True:
				# pylint: disable=maybe-no-member
				response = self.service.files().list(q = sub_q_files,
												spaces = 'drive',
												fields = 'nextPageToken, '
													   f'files({",".join(fields)})',
												pageToken = page_token,
												**self.common_paramters,
												**self.search_paramters
												).execute()
				ret = response.get('files')
				if ret: 
					for file in ret:
						if exclude_dot_files and file.get("name").startswith("."):
							continue
						files.append(file)
				page_token = response.get('nextPageToken', None)
				if page_token is None:
					break
			for file in files:
				file["path"] = path
			page_token = None
			while True:
				# pylint: disable=maybe-no-member
				response = self.service.files().list(q = sub_q_folders,
												spaces = 'drive',
												fields = 'nextPageToken, '
													   f'files({",".join(fields)})',
												pageToken = page_token,
												**self.common_paramters,
												**self.search_paramters
												).execute()
				ret = response.get('files')
				if ret:
					if do_debug:
						logger.info(f"Extending folders {pprint.pformat(folders)} with ret {pprint.pformat(ret)}")
					for folder in ret:
						if exclude_dot_folders and folder.get("name").startswith("."):
							continue
						folders.append(folder)
					if do_debug:
						logger.info(f"To {pprint.pformat(folders)}")
				page_token = response.get('nextPageToken', None)
				if page_token is None:
					break
			recurse_folders = folders.copy()
			for folder in recurse_folders:
				if not folder:
					continue
				if do_debug:
					logger.info(f"RECURSING FOLDER: {pprint.pformat(folder)}, ({pprint.pformat(recurse_folders)})")
				sub_folder_id = folder.get("id")
				sub_path = path.copy()
				sub_path.append(folder)
				sub_files, sub_folders = self.recursive_search_worker(sub_folder_id, q_files, q_folders, fields=fields, path = sub_path, do_debug = do_debug)
				files.extend(sub_files)
				folders.extend(sub_folders)
	
		except googleapiclient.errors.HttpError as error:
			logger.error(f'An error occurred: {error}')
			files = list()
			folders = list()
	
		return files, folders
		

	def download_file(self, file_id, do_debug = False):
		"""Download a file
		Args:
			real_file_id - ID of the file to download
		Returns : IO object with location.
		"""
		file = None
		try:
			# pylint: disable=maybe-no-member
			request = self.service.files().get_media(fileId = file_id)
			file = io.BytesIO()
			downloader = googleapiclient.http.MediaIoBaseDownload(file, request)
			done = False
			while done is False:
				status, done = downloader.next_chunk()
				if do_debug:
					logger.info(f'Download {int(status.progress() * 100)}% of "{file_id}"')
	
		except googleapiclient.errors.HttpError as error:
			logger.error(F'An error occurred: {error}')
			file = None
	
		return file and file.getvalue() or None
	
	
	def create_folder(self, file_metadata = dict()):
		"""Create new folder.
		Args:
			file_metadata - the data for the new folder such as parent ids and name 
		Returns : Id's of the new folder
		"""
		try:
			file_metadata["mimeType"] = FOLDER_MIMETYPE
			# pylint: disable=maybe-no-member
			file = self.service.files().create(body = file_metadata, fields = 'id', **self.common_paramters).execute()
			logger.info(F'Folder ID: "{file.get("id")}".')
			return file.get('id')
		except googleapiclient.errors.HttpError as error:
			logger.error(F'An error occurred: {error}')
		return None

	def _upload_media(self, media, file_metadata = dict(), do_debug = False):
		"""Upload new file.
		Args:
			media - the media to upload
		Returns : Id's of the file uploaded
		"""
		file = None
		try:
			# pylint: disable=maybe-no-member
			file = self.service.files().create(body = file_metadata, media_body = media, fields = 'id', **self.common_paramters).execute()
			if do_debug:
				logger.info(F'File upload resulted in ID: {file.get("id")}')
		except googleapiclient.errors.HttpError as error:
			logger.error(F'An error occurred during file upload: {error}')
			file = None
		return file and file.get('id') or None


	def upload_file(self, file_path, file_metadata = dict(), mimetype = default_mimetype):
		"""Upload new file.
		Args:
			file_path - The local path of the file
			file_metadata - The metadata to give the file (NOTE: In no file name is present, the file_path is used)
			mimetype - The file mimetype
		Returns : Id's of the file uploaded
		"""
		if not file_metadata.get('name'):
			file_metadata['name'] = os.path.basename(file_path)
		media = MediaFileUpload(file_path, mimetype = mimetype)
		return self._upload_media(media = media, file_metadata = file_metadata)
				
	def upload_bytes(self, bytes, file_metadata = dict(), mimetype = default_mimetype):
		"""Upload new bytes object.
		Args:
			bytes - The bytes that make up the file body
			file_metadata - The metadata to give the file (NOTE: Needs to have the file's name)
			mimetype - The file mimetype
		Returns : Id's of the file uploaded
		"""
		if not file_metadata.get('name'):
			return None, "No filename provided"
		media = MediaIoBaseUpload(io.BytesIO(bytes), mimetype = mimetype)
		return self._upload_media(media, file_metadata = file_metadata)

	def update_file_metadata(self, id, file_metadata):
		"""Update file metadata.
		Args:
			id - The id of the file to update
			file_metadata - The metadata to give the file
		Returns : Id's of the file that was updated
		"""
		file = None
		try:
			file = self.service.files().update(fileId=id, body = file_metadata, fields = 'id', **self.common_paramters).execute()
		except googleapiclient.errors.HttpError as error:
			logger.info(F'An error occurred during file update: {error}')
			file = None
		return file and file.get('id') or None

	def file_exists(self, id):
		"""Check if file exists
		Args:
			id - The id of the file to look for
		Returns : True if file exists
		"""
		try:
			ret = self.service.files().get(fileId = id, **self.common_paramters).execute()
			return ret and True or False
		except googleapiclient.errors.HttpError as error:
			logger.info(F'An error occurred during file exist check: {error}')
		return False

	def file_meta(self, id):
		"""Return file meta
		Args:
			id - The id of the file to look for
		Returns : Meta for the file
		"""
		try:
			ret = self.service.files().get(fileId = id, **self.common_paramters).execute()
			return ret or None
		except googleapiclient.errors.HttpError as error:
			logger.info(F'An error occurred during file meta: {error}')
		return False

def create_service_account(project_id, name, display_name):
	"""Creates a service account."""

	credentials = service_account.Credentials.from_service_account_file(
		filename = os.environ['GOOGLE_APPLICATION_CREDENTIALS'],
		scopes = ['https://www.googleapis.com/auth/cloud-platform'])

	service = googleapiclient.discovery.build(
		'iam', 'v1', credentials = credentials)

	my_service_account = service.projects().serviceAccounts().create(
		name = 'projects/' + project_id,
		body = {
			'accountId': name,
			'serviceAccount': {
				'displayName': display_name
			}
		}).execute()

	logger.error('Created service account: ' + my_service_account['email'])
	return my_service_account



from dotenv import dotenv_values
from octomy.cad import openscad
import datetime
import logging
import os
import pprint
import pytest
import re
import shutil

logger = logging.getLogger(__name__)




here = os.path.dirname(__file__)
data_dir = os.path.join(here, "data")
env = dict()
env_re = re.compile(r'export (?P<key>.*)=(?P<value>.*)')


def parse_env(here):
	env_file = os.path.join(here, ".env")
	return dotenv_values(env_file)

def parse_env_old(here):
	env_file = os.path.join(here, ".env")
	out = dict()
	if os.path.isfile(env_file):
		with open(env_file, "r") as f:
			data = f.read()
			#logger.info(f"data: {data}")
			for m in env_re.finditer(data):
				gd = m.groupdict()
				#logger.info(f"gd: {gd}")
				out[gd.get('key')] = gd.get('value')
				#logger.info(f"line: {gd.get('key')} = {gd.get('value')}")
	else:
		logger.warning(f"Could not find env file: {env_file}")
	return out


def get_env(key, fb=None):
	global env
	if env == {}:
		env = parse_env(here)
	ret = env.get(key) or os.environ.get(key) or fb
	logger.info(f"GOT {key}={ret}")
	return ret


def remove_dir(dir):
	logger.info(f"Removing dir '{dir}'")
	if os.path.isdir(dir):
		shutil.rmtree(dir, ignore_errors=True)
	assert(not os.path.isdir(dir))

"""
@pytest.fixture(scope='function')
def git_repo(request):
	logger.info(f"here:{here}, project_folder:{git_folder}")
	git_data = {
		  "path": "github.com/arribatec-cloud/small-test-repo.git"
		, "username": get_env("TEST_GIT_USERNAME", None)
		, "password": get_env("TEST_GIT_PASSWORD", None)
		, "do_debug": True
	}
	logger.info("creating client with:")
	logger.info(pprint.pformat(git_data))
	remove_dir(git_folder)
	repo = git.Client(**git_data)
	logger.info("creating client DONE")
	assert(repo)
	repo.state()
	def teardown():
		logger.info("TEARDONE")
		repo = None
		remove_dir(git_folder)
	request.addfinalizer(teardown)
	return repo
"""



from fastapi import APIRouter, Depends, Request, HTTPException, status
from functools import lru_cache
from pydantic import BaseModel, StringConstraints
import pydantic.version
from typing import List, Optional, Annotated, Any, Dict
import datetime
import logging
import pprint
import uuid
import json


logger = logging.getLogger(__name__)
logger.info(f" # # # # # # PYDANTIC VERSION: {pydantic.version.version_info()}")


some_time = datetime.datetime(2024,1,1,0,0,0)


from pydantic import (
	PlainSerializer,
	TypeAdapter,
	WithJsonSchema,
)


JSONSerializableDict = Annotated[
	dict,
	PlainSerializer(lambda x: json.dumps(x) if isinstance(x, str) else x, return_type=str),
	WithJsonSchema({'type': 'string'}, mode='serialization'),
]


class NewPassword(BaseModel):
	new_password: str


class GeneralID(BaseModel):
	id: uuid.UUID


class UserGroupRelationStump(BaseModel):
	group_id: uuid.UUID
	user_id: uuid.UUID
	
	
class UserGroupRelation(UserGroupRelationStump):
	created_at: datetime.datetime
	
	
class UserStump(BaseModel):
	name: Annotated[str, StringConstraints(max_length=1023)] | None = None
	email: Annotated[str, StringConstraints(max_length=1023)] | None = None
	enabled: bool = False
	super: bool = False
	data: JSONSerializableDict | None = dict()


class User(UserStump):
	id: uuid.UUID
	updated_at:datetime.datetime
	password_changed_at:datetime.datetime | None = None
	login_at:datetime.datetime | None = None
	created_at: datetime.datetime | None = None


	def get_id(self, *args, **kvargs):
		do_debug = False
		if do_debug:
			logger.info(f"user.get_id args='{pprint.pformat(args)}', kvargs='{pprint.pformat(kvargs)}'")
		return id


	def can(self, *args, **kvargs):
		do_debug = False
		if do_debug:
			logger.info(f"user.can args='{pprint.pformat(args)}', kvargs='{pprint.pformat(kvargs)}'")
		return True


class UserDatabase(User):
	password_hash: Annotated[str, StringConstraints(max_length=1023)]


class GroupStump(BaseModel):
	name: Annotated[str, StringConstraints(max_length=1023)] | None
	description: str | None = None
	enabled: bool = False
	data: JSONSerializableDict | None = dict()


class Group(GroupStump):
	id: uuid.UUID
	created_at: datetime.datetime
	updated_at:datetime.datetime


class GrantType(BaseModel):
	key: Annotated[str, StringConstraints(max_length=1023)]
	description: Annotated[str, StringConstraints(max_length=1023)]


class GrantStump(BaseModel):
	key: str
	#key: Annotated[str, StringConstraints(max_length=1023)]


class Grant(GrantStump):
	group_id: uuid.UUID
	created_at: datetime.datetime | None


class UserGrant(BaseModel):
	key: str
	created_at: datetime.datetime | None


@lru_cache()
def get_dummy_user() -> User:
	return User(
		  id="69696969-6969-6969-6969-696969696969"
		, name="Dummy User"
		, email="dummy@example.com"
		, password_hash="dlghjlkszdfjglkdfjglkdfjglkjdfgkljdfgkljsdfgjhayhiyhpiosdvkældfmgjklhwerfuwopdlf"
		, enabled=True
		, super=True
		, data={"data":"lol"}
		, created_at=some_time
		, password_changed_at=some_time
		, login_at=some_time
		, updated_at=some_time
	)

@lru_cache()
def get_current_user() -> User:
	current_user = get_dummy_user()
	return current_user


class AccessContext(BaseModel):
	current_user: User | None
	status_code: int = status.HTTP_500_INTERNAL_SERVER_ERROR
	status_message:str = "Unknown"

	def get_id(self, *args, **kvargs):
		do_debug = False
		if do_debug:
			logger.info(f"AccessContext.get_id args='{pprint.pformat(args)}', kvargs='{pprint.pformat(kvargs)}'")
		if self.current_user:
			return self.current_user.get_id()
		return None

	def can(self, *args, **kvargs):
		do_debug = False
		if do_debug:
			logger.info(f"AccessContext.can args='{pprint.pformat(args)}', kvargs='{pprint.pformat(kvargs)}'")
		if self.current_user:
			return self.current_user.can(*args, **kvargs)
		return False


#@lru_cache()
def dummy_access_context(required_scopes: List[str] | None) -> AccessContext | None:
	return AccessContext(current_user = get_dummy_user(), status_code=status.HTTP_200_OK, status_message="OK")


def get_access_context(required_scopes: Optional[List[str]] = None) -> AccessContext | None:
	user:User = get_current_user()
	if not user:
		return AccessContext(current_user=None, status_code=status.HTTP_401_UNAUTHORIZED, status_message="Not authenticated")
	user_scopes = ["overwatch.access.profile.view"]  # Dummy scopes
	if not set(required_scopes or list()).issubset(set(user_scopes)):
		return AccessContext(current_user=user, status_code=status.HTTP_403_FORBIDDEN, status_message="Insufficient privileges")
	return AccessContext(current_user=user, status_code=status.HTTP_200_OK, status_message="OK")


async def access_context_dependency(required_scopes: Optional[List[str]] = None) -> AccessContext | None:
	#return dummy_access_context(required_scopes)
	async def annotated():
		return get_access_context(required_scopes)
	return annotated


def is_safe_url(request, target):
	ref_url = urlparse(request.host_url)
	test_url = urlparse(urljoin(request.host_url, target))
	return test_url.scheme in ("http", "https") and ref_url.netloc == test_url.netloc


'''



def verify_access_email(email, grant_keys, match_all=True):
	ret = False
	if email:
		db = get_access_db(current_app.config)
		if db.user_is_super(email) > 0:
			ret = True
		else:
			num = db.user_can(email, tuple(grant_keys))
			total = len(grant_keys)
			if match_all:
				# All must match
				if num == total:
					ret = True
			else:
				# At least one must match
				if num > 0:
					ret = True
	return ret


def verify_access_user(user, grant_keys, match_all=True):
	if user:
		return verify_access_email(user.id, grant_keys, match_all)
	return False


def verify_access(grant_keys, match_all=True):
	user = flask_login.current_user
	return verify_access_user(user, grant_keys, match_all)




##################


from functools import wraps

from flask import session, redirect, url_for, request, current_app

from ..common import abortj

from fk.api.access import FlaskLoginUser, flask_login_manager, verify_access, verify_access_user, verify_access_email
from fk.api.access.db import get_access_db

import flask_login
from urllib.parse import urlparse, urljoin


import pprint
import logging

logger = logging.getLogger(__name__)

overwatch_auth = flask_login_manager


overwatch_login_required = flask_login.login_required


def overwatch_login_required_old(func):
    # logger.info("overwatch_login_required")

    @flask_login.login_required
    def wrapper(*args, **kwargs):
        # logger.info("overwatch_login_required.wrapper")
        return func(*args, **kwargs)

    wrapper.__name__ = func.__name__
    return wrapper


def overwatch_verify_access(grant_keys, match_all=True, do_debug=False):
    def decorator(f):
        # logger.info("overwatch_verify_access.decorator")

        @wraps(f)
        def wrapper(*args, **kwargs):
            user = flask_login.current_user
            if verify_access_user(user, grant_keys, match_all):
                if do_debug:
                    logger.info(f"overwatch_verify_access access granted; Needed {'all' if match_all else 'any'} of [{','.join(grant_keys)}]")
                return f(*args, **kwargs)
            if do_debug:
                logger.warning(f"overwatch_verify_access access blocked; Needed {'all' if match_all else 'any'} of [{','.join(grant_keys)}], coming from {request.referrer}")
            url = request.headers.get("Referer")
            if not url:
                url = url_for("access_bp.hall", next=request.args.get("next"))
            return redirect(url)

        return wrapper

    return decorator


def is_safe_url(target):
    ref_url = urlparse(request.host_url)
    test_url = urlparse(urljoin(request.host_url, target))
    return test_url.scheme in ("http", "https") and ref_url.netloc == test_url.netloc


##########################################################################


@flask_login_manager.unauthorized_handler
def unauthorized_handler():
    # logger.info("unauthorized_handler")
    return redirect(url_for("access_bp.login", next=request.args.get("next")))


@flask_login_manager.user_loader
def user_loader(email):
    user = None
    if email:
        authenticate = False
        db = get_access_db(current_app.config)
        db_user = db.get_user_by_email(email)
        if db_user:
            # logger.info(f"FOUND USER IN DB")
            authenticate = True
        if authenticate:
            user = FlaskLoginUser()
            user.id = email
    # logger.info(f"user_loader for {email} returns {user}")
    return user


@flask_login_manager.request_loader
def request_loader(request):
    email = request.form.get("email")
    user = user_loader(email)
    # logger.info(f"request_loader for {email} returns {user}")
    return user







# BASIC



from functools import wraps

from flask import session, redirect, url_for, request, current_app

from flask_httpauth import HTTPBasicAuth
from werkzeug.security import generate_password_hash, check_password_hash

from ..common import abortj

from fk.api.access.db import get_access_db

import pprint
import logging

logger = logging.getLogger(__name__)


overwatch_auth = HTTPBasicAuth()

users_cache = {}


def get_users():
    if not users_cache:
        users_cache = {current_app.config.get("overwatch-user", "no-user"): current_app.config.get("overwatch-password", None)}
        db = get_access_db(current_app.config)
        db_users = db.get_users()
        for user in db_users:
            username = user.get("email")
            password = user.get("password")
            users_cache[username] = password

    return users_cache


def overwatch_login_required(func):
    @overwatch_auth.login_required
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    wrapper.__name__ = func.__name__
    return wrapper


@overwatch_auth.verify_password
def verify_password_basic(username, password):
    users = {current_app.config.get("overwatch-user", "no-user"): current_app.config.get("overwatch-password", None)}
    if username in users and users.get(username, None):
        ret = check_password_hash(generate_password_hash(users.get(username)), password)
        if not ret:
            logger.warning(f"Overwatch verify auth failed")
        return ret
    return False


@overwatch_auth.verify_password
def verify_password(username, password):
    users = get_users()
    if username in users and users.get(username, None):
        ret = check_password_hash(generate_password_hash(users.get(username)), password)
        if not ret:
            logger.warning(f"Overwatch verify auth failed")
            return ret
        return False


def verify_access_worker(username, grant_keys, match_all=True):
    ret = False
    db = get_access_db(current_app.config)
    num = db.user_can(username, tuple(grant_keys))
    total = len(grant_keys)
    if match_all:
        # All must match
        if num == total:
            ret = True
    else:
        # At least one must match
        if num > 0:
            ret = True
    if not ret:
        logger.warning(f"Overwatch verify access failed with {num} of {total} matched, and match_all={match_all}")
    return ret


def overwatch_verify_access(grant_keys, match_all=True):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            username = overwatch_auth.current_user()
            logger.info(f"overwatch_verify_access(username={username}, grant_keys={pprint.pformat(grant_keys)}, match_all={match_all})")
            if not verify_access_worker(username, grant_keys, match_all):
                logger.warning(f"overwatch_verify_access block access. Need {'all' if match_all else 'any'} of {pprint.pformat(grant_keys)}")
                abortj(401, "Access not granted")
            return f(*args, **kwargs)

        return wrapper

    return decorator


##########################################################################


@login_manager.unauthorized_handler
def unauthorized_handler():
    return "Unauthorized Biaaach"


'''



from octomy.utils import get_package_relative_dir, get_package_root_dir, tree
import logging
import octomy.db
import os
import pprint

logger = logging.getLogger(__name__)

# fmt: off
configs = {
	"empty:":(
		{}
		,""
		,""
	)
	, "hello:":(
		{
			"db-hostname":"hello.com"
			,"db-port":1234
			,"db-username":"arnold"
			,"db-password":"secret123"
			,"db-database":"mydb"
		}
		,"F12F52B73358C297F47A80768ABDFADF20D021F6A20E9929178908F981B75FA1"
		,"postgres://arnold:secret123@hello.com:1234/mydb"
	)
}
# fmt: on


def _test_redact():
	min = 10
	base = "lorem ipsum dolor sit amet alla balla tralla lalla"
	l = len(base)
	for i in range(0, min*2):
		sub = ""
		for j in range(0, i):
			sub += base[ j % l ]
		redacted = octomy.utils.redact(sub, min)
		logger.info(f"{sub} --> {redacted}")

def _test_db_get_config_hash():

	for name, pack in configs.items():
		logger.info(f"NAME:{name}")
		config, expected, _ = pack
		logger.info(f"config:{config}")
		logger.info(f"expected:{expected}")
		actual = octomy.db.get_config_hash(config)
		logger.info(f"actual:{actual}")
		assert actual == expected
	return True


def _test_db_uri_to_and_from_config():
	for name, pack in configs.items():
		logger.info(f"NAME:{name}")
		expected_config, _, expected_uri = pack
		logger.info(f"expected_config:{expected_config}")
		logger.info(f"expected_uri:   {expected_uri}")
		actual_uri, actual_uri_err = octomy.db.db_uri_from_conf(expected_config, do_online=False)
		actual_config = octomy.db.db_uri_to_config(expected_uri, do_online=False)
		if None == actual_uri:
			logger.info(f"actual_uri_err={actual_uri_err}")
		else:
			assert actual_uri == expected_uri
			assert actual_config == expected_config

def _test_tree():
	path = get_package_relative_dir(True)
	logger.info(f"relpath: {path}")
	listing = tree(path.parent)
	logger.info(f"listing:\n{listing}")
	#tree()


def _test_stripes():
	#logger.info(stripes())
	matrix = [
		{
			  "expected_result": '\x1b[94m###   ###   ###   ###   ###   ###   ###   ###   ###   ###   ###   ###   ###   #\x1b[0m\n\x1b[94m##   ###   ###   ###   ###   ###   ###   ###   ###   ###   ###   ###   ###   ##\x1b[0m\n\x1b[94m#   ###   ###   ###   ###   ###   ###   ###   ###   ###   ###   ###   ###   ###\x1b[0m\n'
			, "params": {'cols':79, 'rows':3, 'on':3, 'off':3, 'on_char':"#", 'off_char':" ", 'offset':0, 'sheer':1}
		}
		, {
			  "expected_result": '\x1b[94m*-****-****-****-****-****-****-****-***\x1b[0m\n\x1b[94m-****-****-****-****-****-****-****-****\x1b[0m\n'
			, "params": {'cols':40, 'rows':2, 'on':4, 'off':1, 'on_char':"*", 'off_char':"-", 'offset':3, 'sheer':1}
		}
		, {
			  "expected_result": '\x1b[94m!%%!!%%!!%\x1b[0m\n\x1b[94m%%!!%%!!%%\x1b[0m\n\x1b[94m%!!%%!!%%!\x1b[0m\n\x1b[94m!!%%!!%%!!\x1b[0m\n\x1b[94m!%%!!%%!!%\x1b[0m\n\x1b[94m%%!!%%!!%%\x1b[0m\n\x1b[94m%!!%%!!%%!\x1b[0m\n\x1b[94m!!%%!!%%!!\x1b[0m\n\x1b[94m!%%!!%%!!%\x1b[0m\n\x1b[94m%%!!%%!!%%\x1b[0m\n'
			, "params": {'cols':10, 'rows':10, 'on':2, 'off':2, 'on_char':"%", 'off_char':"!", 'offset':3, 'sheer':1}
		}
	]
	for case in matrix:
		expected_result = case.get("expected_result")
		params = case.get("params")
		actual_result = octomy.utils.stripes(**params)
		logger.info(f"CASE expected_result={expected_result} params={params}")
		logger.info("EXPECTED RESULT:")
		logger.info(pprint.pformat(expected_result))
		logger.info("ACTUAL RESULT:")
		logger.info(pprint.pformat(actual_result))
		assert(actual_result == expected_result)


# fmt: off
query_key_test_cases = {
	"empty:":(
		  ""
		, None
	)
	, "long:":(
		  "octomy.batch.sub.dub.ding"
		, "octomy/batch/sub/dub/sql/ding.sql"
	)
}
# fmt: on

def _test_query_key_to_sql_path_relative():
	for name, pack in query_key_test_cases.items():
		query_key, expected = pack
		actual = octomy.db.query_key_to_sql_path(query_key = query_key, relative = True, do_debug = True)
		logger.info(f"NAME:{name}")
		logger.info(f"  query_key:{query_key}")
		logger.info(f"   expected:{expected}")
		logger.info(f"     actual:{actual}")
		assert actual == expected


def _test_get_package_root_dir():
	dir = get_package_root_dir()
	logger.info(f"get_package_root_dir:{dir}")


def _test_combined():
	package_dir = get_package_root_dir('octomy.utils')
	logger.info(f"  get_package_root_dir:{package_dir}")
	assert bool(package_dir)
	package_dir = package_dir.parent
	query_key = "octomy.db.get_now"
	logger.info(f"             query_key:{query_key}")
	sql_path = octomy.db.query_key_to_sql_path(query_key = query_key, relative = True, do_debug = True)
	logger.info(f"              sql_path:{sql_path}")
	assert sql_path == "octomy/db/sql/get_now.sql"
	canonical_path = os.path.join(package_dir, sql_path)
	logger.info(f"        canonical_path:{canonical_path}")
	assert bool(canonical_path)
	assert os.path.exists(canonical_path)




def _test_xray():
	
	bob = {
		  "alla":  [(1, 3), (3, 3), (1, "bob"), (3, 1), (5, 0), (4, 10)]
		, "balla":  ["lo", "bo", "ho"]
		, "sjokki": { "lol":123445, "bob":3452345.3455, "foo":"bar"}
		, "hehe": None
		, "hoho": False
	}
	
	ret = octomy.utils.xray(bob)


def _test_random_token():
	for i in range(0, 100):
		logger.info(octomy.utils.random_token(length=10))



#   __ __                 ___        _            
#  |  \  \ ___  ___  ___ / __> ___ _| |_ _ _  ___ 
#  |     |/ ._>/ . |<_> |\__ \/ ._> | | | | || . \
#  |_|_|_|\___.\_. |<___|<___/\___. |_| `___||  _/
#              <___'                         |_|  

from setuptools import find_namespace_packages
import logging
import os
import pprint
import re
import sys
import inspect
import warnings

log_setup_once = False


def setup_logging(name=None, log_level=logging.INFO):
	"""Set up logging."""
	global log_setup_once
	if not log_setup_once:
		log_setup_once = True
		logging.basicConfig(level=log_level)
		fmt = "%(asctime)s %(levelname)s (%(threadName)s) [%(name)s:%(lineno)s::%(funcName)s()] - %(message)s"
		colorfmt = "%(log_color)s{}%(reset)s".format(fmt)
		datefmt = "%Y-%m-%d %H:%M:%S"

		# Suppress overly verbose output that isn't helpful from some libraries we depend on
		for key in ["requests", "tensorboard", "urllib3", "aiohttp.access", "uamqp", "sqlalchemy", "sqlalchemy.engine.base", "matplotlib.font_manager"]:
			logging.getLogger(key).setLevel(logging.WARNING)

		# Enable debug logging for some insteresting libraries (for development)
		logging.getLogger("fk").setLevel(logging.DEBUG)

		try:
			from colorlog import ColoredFormatter

			logging.getLogger().handlers[0].setFormatter(ColoredFormatter(colorfmt, datefmt=datefmt, reset=True, log_colors={"DEBUG": "cyan", "INFO": "green", "WARNING": "yellow", "ERROR": "red", "CRITICAL": "red"}))
		except ImportError:
			# Ignore failures in loading color coded logs
			pass

		# Set this log level as default
		logger = logging.getLogger("")
		logger.setLevel(log_level)
	if not name:
		name = inspect.stack()[1][1]
	logger = logging.getLogger(name)
	try:
		import colored_traceback.auto
	except ImportError:
		pass

	return logger


logger = setup_logging(__name__)


def read_file(cwd:str, fname:str, strip:bool = True):
	fn = os.path.realpath(os.path.join(cwd, fname))
	data = ""
	if os.path.exists(fn):
		with open(fn) as f:
			data = f.read()
			data = data.strip() if strip else data
			# logger.info(f"Got data '{data}' from '{fn}'")
	else:
		logger.error(f"Could not find file {fn}")
		logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
	return data


def write_file(cwd:str, fname:str, data:str, do_overwrite:bool = False):
	fn = os.path.realpath(os.path.join(cwd, fname))
	if not os.path.exists(fn) or do_overwrite:
		with open(fn, "w") as f:
			f.write(data)
	else:
		logger.warning(f"File {fn} already exists")
		logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
	return data


def remove_comment(line, sep="#"):
	i = line.find(sep)
	if i >= 0:
		line = line[:i]
	return line.strip()


def read_requirements_file(cwd:str, fname:str, do_strip:bool = True, do_debug:bool = False):
	fn = os.path.realpath(os.path.join(cwd, fname))
	if do_debug:
		logger.info(f"Reading requirements from {fn} with do_strip = {do_strip}")
	lines = []
	if os.path.exists(fn):
		with open(fn) as f:
			for r in f.readlines():
				r = r.strip()
				if len(r) < 1:
					continue
				r = remove_comment(r)
				if len(r) < 1:
					continue
				lines.append(r)
	else:
		logger.error(f"Could not find requirements file {fn}")
		logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
	# logger.warning(f"Full content of '{fname}' was: \n{lines}")
	if not do_strip:
		return lines
	out = []
	for line in lines:
		if line and not line.startswith("-"):
			out.append(line)
	return out


def debug_repo(repo):
	if not repo:
		logger.info(f"No repo")
		return
	logger.info(f"Repository head commit: {repo.head.commit}")
	logger.info(f"Found {len(repo.branches)} branches:")
	for branch in repo.branches:
		logger.info(f" + {branch}({branch.commit})")
	remote = repo.remote()
	logger.info(f"Found {len(remote.refs)} remote refs:")
	for ref in remote.refs:
		logger.info(f" + {ref}({ref.commit})")


def get_git_branch_from_env():
	branch_env = "FK_GIT_ACTUAL_BRANCH"
	branch = os.environ.get(branch_env, None)
	if branch is not None:
		logger.info(f"Using {branch_env} = {branch} from environment")
	else:
		logger.info(f"No value for {branch_env} found")
	return branch

def get_license_name(cwd:str, fname:str, do_debug:bool = False):
	fn = os.path.realpath(os.path.join(cwd, fname))
	if do_debug:
		logger.info(f"Reading license from {fn}")
	if os.path.exists(fn):
		with open(fn) as f:
			for r in f.readlines():
				r = r.strip()
				if len(r) < 1:
					continue
				# Return first non-empty line
				return r
	else:
		logger.error(f"Could not find license file {fn}")
		logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
	return "Proprietary" # Fall back to something safe


def generate_version_string(version:str = None, branch:str = None):
	version = read_file(version_file) if version is None else version
	branch = get_git_branch_from_env() if branch is None else branch
	full_version = ""

	if branch == "production":
		full_version = version
	elif branch == "beta":
		full_version = f"{version}-beta"
	elif branch.startswith("stage-") and len(branch) > 6:
		full_version = f"{version}-{branch[6:]}"
	else:
		full_version = f"{version}-test-{branch.replace(' ','_').replace('	','_')}"
	logger.info(f"Using full version = {full_version}")
	return full_version


def generate_development_status(cwd:str, version:str = None, branch:str = None):
	version = read_file(cwd, version_file) if version is None else version
	branch = get_git_branch_from_env() if branch is None else branch
	development_status = ""
	# Calculate the development status based on current branch and bare version
	if branch == release_branch:
		development_status = "Development Status :: 5 - Production/Stable"
	elif branch == beta_branch:
		development_status = "Development Status :: 4 - Beta"
	elif "feature-" in branch:
		development_status = "Development Status :: 3 - Alpha"
	else:
		development_status = "Development Status :: 1 - Planning"
	return development_status


def get_development_status():
	# Not viable
	# return generate_development_status();
	return "Development Status :: 1 - Planning"

def find_packages(package_dir:str, modules:list, do_debug:bool = False):
	#return find_namespace_packages(where=package_dir, include=[module + ".*" for module in modules])
	includes = list()
	for module in modules:
		if do_debug:
			logger.info(f"Searching {module} for includes..")
		includes.append(module + ".*")
	if do_debug:
		logger.info(f"Includes found: {pprint.pformat(includes)}")
	found_packages = find_namespace_packages(where=package_dir, include=includes)
	if do_debug:
		logger.info(f"Packages found: {pprint.pformat(found_packages)}")
	return found_packages


# Function to recursively list all files in a directory
def list_files(directory:str, base:str):
	paths = list()
	for (path, directories, filenames) in os.walk(directory):
		for filename in filenames:
			paths.append(os.path.relpath(os.path.join(path, filename), base))
	return paths

def get_package_data(extensions:list, modules:list, package_dir:str, do_debug):
	out = dict()
	if do_debug:
		logger.info(f"get_package_data(extensions={extensions}, modules={modules}, package_dir={package_dir})")
	data_filters = list()
	for extension in extensions:
		data_filters.append(re.compile(fr'.*/{extension}(?:/{extension})*/[^/]+\.{extension}$'))
	for module in modules:
		modules_data_files = list_files(module, os.path.join(package_dir))
		data_files = list()
		for data_file in modules_data_files:
			for data_filter in data_filters:
				if data_filter.match(data_file):
					data_files.append(data_file)
					p = os.path.dirname(data_file)
					f = os.path.basename(data_file)
					m = ".".join(p.split("/"))
					ms = out.get(m, list())
					ms.append(f)
					out[m] = ms
					#logger.info(f"########## FOUND: {data_file}    (p={p}, f={f}, m={m})")
					break

	logger.info("Datafiles:---")
	logger.info(pprint.pformat(out))
	logger.info("-------------")
	return out

def clean_module_folder_name(dash_name):
	return dash_name.replace('-', '_')

# From https://pypi.org/pypi?%3Aaction=list_classifiers
def get_classifiers(python_version:str):
	return [
		  get_development_status()
		, "Intended Audience :: Developers"
		, "Intended Audience :: Information Technology"
		, "Intended Audience :: Science/Research"
		, "Intended Audience :: Other Audience"
		, "Topic :: Utilities"
		, "Natural Language :: English"
		, "Operating System :: POSIX :: Linux"
		, "Programming Language :: Python :: " + python_version
		, "Topic :: Other/Nonlisted Topic"
	]

def megasetup(source_data:dict):
	do_debug = source_data.get("debug", False)
	base_name = source_data.get("base_name")
	data_extensions = source_data.get("data_extensions", ['sql'])
	modules = set(source_data.get("modules", set()))
	additional_modules = source_data.get("additional_modules", list())
	group_base_name = source_data.get("group_base_name")
	package_dir = source_data.get("package_dir", "src")
	cwd = source_data.get("cwd", os.path.realpath(os.path.dirname(os.path.abspath(__file__))) )
	has_cli = source_data.get("has_cli", False)
	executable_name = source_data.get("executable_name")
	executable_package = source_data.get("executable_package")
	executable_package_path = source_data.get("executable_package_path")
	python_version = source_data.get("python_version", "3.10")
	version_file = "./VERSION"
	readme_file = "./README.md"
	license_file = "./LICENSE"
	
	if group_base_name:
		if len(modules) < 1:
			modules.add(clean_module_folder_name(group_base_name))
	modules.update(additional_modules)
	modules = list(modules)
	url = f"https://gitlab.com/{group_base_name}/{base_name}"
	scripts = list()
	entry_points = dict()
	if do_debug:
		logger.info(f"MEGASETUP       base_name: {base_name}")
		logger.info(f"                      cwd: {cwd}")
		logger.info(f"          data_extensions: {data_extensions}")
		logger.info(f"          group_base_name: {group_base_name}")
		logger.info(f"                  has_cli: {has_cli}")
		logger.info(f"             license_file: {license_file}")
		logger.info(f"                  modules: {modules}")
		logger.info(f"              package_dir: {package_dir}")
		logger.info(f"           python_version: {python_version}")
		logger.info(f"              readme_file: {readme_file}")
		logger.info(f"                      url: {url}")
		logger.info(f"             version_file: {version_file}")
	if has_cli or executable_name or executable_package or executable_package_path:
		executable_name = executable_name or base_name
		if not executable_package_path:
			executable_package_path=[clean_module_folder_name(group_base_name)]
			if executable_package:
				executable_package_path.append(executable_package)
			executable_package_path = '.'.join(executable_package_path)
		executable = f"{executable_name} = {executable_package_path}.cli:main"
		if executable_name:
			entry_points={"console_scripts": [executable]}

	package_data = get_package_data(data_extensions, modules, package_dir, do_debug)

	defaults_data = {
		  "author_email": "pypi@octomy.org"
		, "author_name": "OctoMY"
		, "beta_branch": "beta"
		, "changelog_file": "./CHANGELOG"
		, "classifiers": get_classifiers(python_version)
		, "entry_points": entry_points
		, "include_package_data": True
		# Allow flexible deps for install
		, "install_requires": read_requirements_file(cwd, "requirements/requirements.in", do_debug=do_debug)
		, "keywords": ["software"]
		, "license_files": (license_file,)
		, "license_name": get_license_name(cwd, license_file, do_debug=do_debug)
		, "long_description": read_file(cwd, readme_file)
		, "long_description_content_type": "text/markdown"
		# We use namespace packages to allow multiple packages to use the same package prefix
		# We omit __init__.py to accomplish this
		# See https://packaging.python.org/en/latest/guides/packaging-namespace-packages/
		, "modules": modules
		# NOTE: "data_files" is deprecated
		# NOTE: "package_data" need to reside inside a package, in other words a directory with __init__.py
		, "package_data": package_data
		, "package_dir": {'': package_dir}
		, "package_name": f"{group_base_name}_{base_name}" # see PEP625 https://peps.python.org/pep-0625/
		, "packages": find_packages(package_dir, modules)
		, "platforms": ["Linux"]
		, "project_urls": {"Docs": f"{url}//-/blob/production/README.md", "Bugs": f"{url}/issues", "C.I.": f"{url}/pipelines"}
		, "python_requires": ">=" + python_version
		, "scripts": scripts
		, "python_version": python_version
		, "release_branch": "development"
		, "setup_requirements": ["pytest-runner", "setuptools_scm"]
		, "short_description": f"{group_base_name}/{base_name}"
#		, "test_suite": os.path.join(package_dir, "tests")
		# Use flexible deps for testing
		, "tests_require": read_requirements_file(cwd, "requirements/test_requirements.in", do_debug=do_debug)
		, "url": url
		, "version_string": read_file(cwd, version_file)
		, "zip_safe": True
	}

	for key, value in defaults_data.items():
		source_data.setdefault(key, value)



	package = {
		  "author": source_data.get("author_name")
		, "author_email": source_data.get("author_email")
		, "classifiers": source_data.get("classifiers")
		, "description": source_data.get("short_description")
		, "entry_points": source_data.get("entry_points")
		, "include_package_data": source_data.get("include_package_data")
		, "install_requires": source_data.get("install_requires")
		, "keywords": ", ".join(source_data.get("keywords", []))
		, "license": source_data.get("license_name")
		, "license_files": source_data.get("license_files")
		, "long_description": source_data.get("long_description")
		, "long_description_content_type": source_data.get("long_description_content_type")
		, "maintainer": source_data.get("author_name")
		, "maintainer_email": source_data.get("author_email")
		, "name": source_data.get("package_name")
		, "namespace_packages": source_data.get("modules")
		, "package_data": source_data.get("package_data")
		, "package_dir": source_data.get("package_dir")
		, "packages": source_data.get("packages", list())
		, "platforms": source_data.get("platforms")
		, "project_urls": source_data.get("project_urls")
		, "python_requires": source_data.get("python_requires")
		, "setup_requires": source_data.get("setup_requirements")
#		, "test_suite": source_data.get("test_suite") # Deprecated
#		, "tests_require": source_data.get("tests_require")
		, "url": source_data.get("url")
		, "version": source_data.get("version_string")
		, "zip_safe": source_data.get("zip_safe")
#		, "scripts": source_data.get("scripts") #Using entry_points instead
	}
	#del package['package_data']
	for key,val in package.items():
		if not val:
			logger.info(f" {key}:{val}  <-- EMPTY ¤ ¤ ¤ ¤ ¤ ¤ ¤ ¤")
		else:
			logger.info(f" {key}:{len(str(val))} bytes")
	if do_debug:
		logger.info("-------------------------------------------------------")
		logger.info("'  setup.py package:")
		logger.info(pprint.pformat(package))
		logger.info("-------------------------------------------------------")
	return package



import logging
import octomy.db
import octomy.web.context
import octomy.cad.parts
import os
import pprint
import pytest
import datetime


logger = logging.getLogger(__name__)

# fmt: off
dummy_config= {
	"db-hostname":"hello.com"
	,"db-port":"1234"
	,"db-username":"arnold"
	,"db-password":"secret123"
	,"db-database":"mydb"
}
# fmt: on



@pytest.mark.asyncio
async def test_parts():
	#parts = octomy.cad.parts.Parts()
	pass




#!/usr/bin/env python
from setuptools import setup
import logging
import os
import glob
import pprint
import sys

source_data = {
	  "base_name": "common"
	, "group_base_name": "octomy"
	, "cwd": os.path.dirname(os.path.abspath(__file__))
	, "debug": True
}

ms = None
sl = None

try:
	#sys.path.insert(0, project_path)
	#print(f"balls:{project_path}")
	from octomy.utils.setup import megasetup, setup_logging
	ms = megasetup
	sl = setup_logging
	project_path = os.path.join(os.path.abspath(os.path.dirname(__file__)))
except Exception as e:
	print(" ")
	print("^^^^^^^^^^^^^^^^^^^^^^^")
	print(f"Megasetup not available: {e}")
	print("^^^^^^^^^^^^^^^^^^^^^^^")
	print(" ")
	print(f" + project_path: {project_path}")
	for parent_path, _, filenames in os.walk(project_path):
		for f in filenames:
			print(f"   found {os.path.join(parent_path, f)}")
	print(" ")
	sys.exit(1)

logger = sl(__name__)


logger.info("source_data")
logger.info(pprint.pformat(source_data))

try:
	package = ms(source_data = source_data)
	logger.info("package:")
	logger.info(pprint.pformat(package))
	logger.info("setup():")
	setup(**package)
except Exception as e:
	logger.error(f"Error during megasetup: {e}")
	logger.exception(e)




import logging
from furl import furl
import fk.utils

logger = logging.getLogger(__name__)

def config_to_url(config, do_redact=False):
	if not config:
		return ""
	db_hostname = config.get("db-hostname", None)
	db_port = config.get("db-port", None)
	db_username = config.get("db-username", None)
	db_password = config.get("db-password", None)
	db_database = config.get("db-database", None)
	if do_redact:
		db_password = fk.utils.redact(db_password)
	f = furl()
	f.set(scheme="postgres", username=db_username, password=db_password, host=db_hostname, port=db_port, path=db_database)
	return f.url


def url_to_config(db_url):
	f = furl(db_url)
	if f.scheme != "postgres":
		logger.warning("db_url URL did not have postgres scheme")
		return {}
	database = (f.path.segments or [""])[0]
	# fmt:off
	config={
	  "db-hostname": f.host
	, "db-port": f.port
	, "db-username": f.username
	, "db-password": f.password
	, "db-database": database
	}
	# fmt:on
	return config


def get_config_hash(config):
	if not config:
		return ""
	key = ""
	for i in ["hostname", "port", "username", "password", "database"]:
		key += fk.utils.hashstr(config.get(f"db-{i}", ""))
	return fk.utils.hashstr(key)



from influxdb import InfluxDBClient
from pprint import pformat
import traceback
import logging

logger = logging.getLogger(__name__)


# Helper class to manage influxdb
class Influx:
	def __init__(self, credentials):
		logger = setup_logging(self.__class__.__name__)
		self.credentials = credentials
		self.client = InfluxDBClient(self.credentials["influx-hostname"], self.credentials["influx-port"], self.credentials["influx-username"], self.credentials["influx-password"], self.credentials["influx-dbname"])
		self.client.create_database(self.credentials["influx-dbname"])
		self.transaction_points = None
		self.transaction_tags = {}
		self.transaction_time_precision = "ms"

	# Begin super basic transaction
	def begin_transaction(self, time_precision="ms"):
		if None is not self.transaction_points:
			logger.error("Another transaction is in progress, aborting")
			return False
		self.transaction_points = []
		self.transaction_tags = {}
		self.transaction_time_precision = time_precision
		return True

	# Add point to super basic transaction
	def add_point_transaction(self, measurement, time, tags={}, fields={}):
		if None is self.transaction_points:
			logger.error("Transaction not started, aborting")
			return False
		# Keep tags in set
		for key, val in tags.items():
			self.transaction_tags[key] = val
		# Keep point
		point_body = {"measurement": measurement, "tags": tags, "time": time, "fields": fields}
		self.transaction_points.append(point_body)
		return True

	# Commit super basic transaction
	def commit_transaction(self):
		ret = False
		if self.transaction_points is not None:
			try:
				ret = self.client.write_points(points=self.transaction_points, time_precision=self.transaction_time_precision, tags=self.transaction_tags)
			except Exception as e:
				logger.error(f" ")
				logger.error(f"#########################################################")
				logger.error(f"## ERROR: Could not commit influx transaction")
				logger.error(f"## Transaction details':")
				logger.error(f"##")
				logger.error(f"## time_precision={self.transaction_time_precision}")
				logger.error(f"## points=")
				logger.error(pformat(self.transaction_points))
				logger.error(f"## tags=")
				logger.error(pformat(self.transaction_tags))
				logger.error(f"##")
				logger.error(f"## {e}:")
				logger.error(f"##")
				traceback.print_exc()
				logger.error(f"##")
				logger.error(f"#########################################################")
				logger.error(f" ")
			self.transaction_points = None
			self.transaction_tags = {}
		return ret

	# Simple single point insertion
	def put_point(self, measurement, time, tags={}, fields={}):
		body = [{"measurement": measurement, "tags": tags, "time": time, "fields": fields}]
		return self.client.write_points(body)

	# Perform query
	def query(self, q):
		return self.client.query(q)
